unit unitkeyedit;

(* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version 
 * 1.1 (the "License"); you may not use this file except in compliance with 
 * the License. You may obtain a copy of the License at 
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Lightning TKeyEdit class.
 *
 * The Initial Developer of the Original Code is
 * Tomsoft.
 * Portions created by Tomsoft are Copyright (C) 2009
 * Tomsoft. All Rights Reserved.
 *
 * Contributor(s):
 *   Tommy Sools <tom@sools.com>
 *
 * ***** END LICENSE BLOCK ***** *)

 interface

uses
  StdCtrls, Messages, Classes, Windows, SysUtils;

type
  TShiftControlEvent = procedure(Sender: TObject; shift: Boolean; control: Boolean) of object;
  TKeyEdit = class(TEdit)
  private
    fVirtualKey: Longint;
    fOnShiftControl: TShiftControlEvent;
    procedure keyEditDblClick(Sender: TObject);
  protected
    procedure Change; override; 
    procedure WMKEYDOWN(var Message: TMessage); message WM_KEYDOWN;
    procedure WMKEYUP(var Message: TMessage); message WM_KEYUP;
  public
    class function translateKey(virtualKey: DWORD): string;
    constructor Create(AOwner: TComponent); override;
    procedure setVirtualKey(virtualKey: Longint);
    function getVirtualKey(): Longint;
  published
    property OnShiftControl: TShiftControlEvent read fOnShiftControl write fOnShiftControl;
  end;

implementation

class function TKeyEdit.translateKey(virtualKey: DWORD): string;
var
  c: Char;
begin
  case virtualKey of
    VK_ESCAPE: Result := 'Escape';
    VK_TAB:    Result := 'Tab';
    VK_BACK:   Result := 'Back';
    VK_RETURN: Result := 'Enter';
    VK_SPACE:  Result := 'Space';
    VK_PAUSE:  Result := 'Pause';
    VK_INSERT: Result := 'Insert';
    VK_HOME:   Result := 'Home';
    VK_END:    Result := 'End';
    VK_NEXT:   Result := 'PageUp';
    VK_PRIOR:  Result := 'PageDn';
    VK_UP:     Result := 'Up';
    VK_DOWN:   Result := 'Down';
    VK_LEFT:   Result := 'Left';
    VK_RIGHT:  Result := 'Right';
    VK_LWIN,
    VK_RWIN:   Result := 'Win';
    VK_NUMPAD0: Result := 'Num0';
    VK_NUMPAD1: Result := 'Num1';
    VK_NUMPAD2: Result := 'Num2';
    VK_NUMPAD3: Result := 'Num3';
    VK_NUMPAD4: Result := 'Num4';
    VK_NUMPAD5: Result := 'Num5';
    VK_NUMPAD6: Result := 'Num6';
    VK_NUMPAD7: Result := 'Num7';
    VK_NUMPAD8: Result := 'Num8';
    VK_NUMPAD9: Result := 'Num9';
    VK_APPS:    Result := 'Menu';
    0,
    VK_SCROLL,
    VK_F1,
    VK_F2,
    VK_F3,
    VK_F4,
    VK_F5,
    VK_F6,
    VK_F7,
    VK_F8,
    VK_F9,
    VK_F10,
    VK_F11,
    VK_F12,
    VK_F13,
    VK_F14,
    VK_F15,
    VK_F16,
    VK_F17,
    VK_F18,
    VK_F19,
    VK_F20,
    VK_F21,
    VK_F22,
    VK_F23,
    VK_F24,
    VK_PRINT,
    VK_DELETE,
    VK_MENU: Result := '';
  else
    c := Chr(MapVirtualKey(virtualKey, 2));
    if c <> #0 then
      Result := c
    else
      Result := '[' + IntToStr(virtualKey) + ']';
  end;
end;

constructor TKeyEdit.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  fVirtualKey := 0;
  @fOnShiftControl := nil;
  OnDblClick := keyEditDblClick;
end;

procedure TKeyEdit.setVirtualKey(virtualKey: Longint);
begin
  fVirtualKey := virtualKey;
  Text := translateKey(virtualKey);
  if Text = '' then
    fVirtualKey := 0;
end;

function TKeyEdit.getVirtualKey(): Longint;
begin
  Result := fVirtualKey;
end;

procedure TKeyEdit.Change();
begin
  // ignore text changes;
  // the OnChange event will be triggered in the WM_KEYDOWN message handler.
end;

procedure TKeyEdit.WMKEYDOWN(var Message: TMessage);
var
  nVirtKey: Longint;
begin
  nVirtKey := Longint(Message.WParam);
  if @OnShiftControl <> nil then
  begin
    OnShiftControl(
      Self,
      HIBYTE(GetKeyState(VK_SHIFT)) > 0,
      HIBYTE(GetKeyState(VK_CONTROL)) > 0
    );
  end;
  if (nVirtKey = VK_SHIFT) or (nVirtKey = VK_CONTROL) then
  begin
    setVirtualKey(0);
    OnChange(Self);
  end
  else if (nVirtKey <> VK_SCROLL) then
  begin
    setVirtualKey(nVirtKey);
    OnChange(Self);
  end;
end;

procedure TKeyEdit.WMKEYUP(var Message: TMessage);
var
  nVirtKey: Longint;
begin
  nVirtKey := Longint(Message.WParam);
  if (getVirtualKey() = 0) and (nVirtKey <> VK_SCROLL) then
  begin
    if @OnShiftControl <> nil then
    begin
      OnShiftControl(
        Self,
        HIBYTE(GetKeyState(VK_SHIFT)) > 0,
        HIBYTE(GetKeyState(VK_CONTROL)) > 0
      );
    end;
  end;
end;

procedure TKeyEdit.keyEditDblClick(Sender: TObject);
begin
  setVirtualKey(0);
  OnChange(Self);
end;

end.
